package org.artifactory.event;

import com.google.common.annotations.VisibleForTesting;
import lombok.Builder;
import lombok.Data;
import org.artifactory.api.context.ContextHelper;
import org.artifactory.common.ConstantValues;
import org.artifactory.schedule.JobCommand;
import org.artifactory.schedule.JobName;
import org.artifactory.schedule.QuartzCommand;
import org.artifactory.schedule.Task;
import org.artifactory.schedule.TaskService;
import org.artifactory.schedule.TaskUser;
import org.artifactory.schedule.TaskUtils;
import org.artifactory.spring.ContextCreationListener;
import org.artifactory.storage.db.event.entity.NodeEventCursorType;
import org.artifactory.storage.db.event.error.service.NodeEventsErrorsStorageService;
import org.artifactory.storage.db.event.model.NodeEventCursor;
import org.artifactory.storage.db.event.service.NodeEventCursorService;
import org.artifactory.storage.event.EventsService;
import org.jfrog.common.ClockUtils;
import org.joda.time.DateTime;
import org.quartz.JobExecutionContext;
import org.quartz.JobExecutionException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.util.Date;

import static java.lang.Math.max;
import static java.lang.Math.min;
import static java.util.concurrent.TimeUnit.HOURS;

@Service
public class EventsLogCleanUpService implements ContextCreationListener {

    private static final Logger log = LoggerFactory.getLogger(EventsLogCleanUpService.class);
    public static final String DELETE_MARKER = "DELETE_MARKER";

    private final TaskService taskService;

    private final EventsService eventsService;

    private final NodeEventCursorService nodeEventCursorService;

    private final NodeEventsErrorsStorageService eventErrorsService;

    @Autowired
    public EventsLogCleanUpService(TaskService taskService, EventsService eventsService,
            NodeEventCursorService nodeEventCursorService, NodeEventsErrorsStorageService eventErrorsService) {
        this.taskService = taskService;
        this.eventsService = eventsService;
        this.nodeEventCursorService = nodeEventCursorService;
        this.eventErrorsService = eventErrorsService;
    }

    @Override
    public void onContextCreated() {
        Task repeatingTask = TaskUtils.createRepeatingTask(EventsLogCleanUpTask.class,
                HOURS.toMillis(ConstantValues.eventsLogCleanupTaskPeriodHours.getLong()),
                30000,
                "Cleanup old events from events log");
        taskService.startTask(repeatingTask, false);
    }

    public void cleanup() {
        log.info("Starting cleanup of old events from event log");
        if (eventsService.getEventsCount() == 0) {
            log.info("No events to cleanup in events log");
            return;
        }
        long eventsDeleteInterval = HOURS.toMillis(ConstantValues.eventsCleanupIntervalHours.getLong());
        final MaxBoundary maxBoundary = getMaxBoundary();
        long deleteFrom = eventsService.getFirstEventTimestamp();
        if (deleteFrom == 0) {
            log.warn("First event is in time 0");
            return;
        }
        log.debug("Delete from {} to {}", deleteFrom, maxBoundary);

        // Check if errors deletion is required
        if (maxBoundary.isDeleteErrors() && deleteFrom < maxBoundary.getTimestamp()) {
            //Delete errors first: to avoid integrity violation of events
            long deleted = eventErrorsService.deleteBeforeWithEventId(maxBoundary.getTimestamp());
            log.debug("{} errors deleted", deleted);
        }

        while (deleteFrom < maxBoundary.getTimestamp()) {
            final long deleteTo = min(deleteFrom + eventsDeleteInterval, maxBoundary.getTimestamp());
            log.trace("Deleting events from {} to {}", deleteFrom, deleteTo);
            //Delete Events
            eventsService.deleteRange(deleteFrom, deleteTo);
            nodeEventCursorService.updateOrInsertCursor(
                    new NodeEventCursor(DELETE_MARKER, deleteTo, NodeEventCursorType.DELETE_MARKER));
            deleteFrom = deleteTo;
            sleep();
        }
        log.info("Cleanup of old events from event log finished");
    }

    @VisibleForTesting
    MaxBoundary getMaxBoundary() {
        long maxBoundaryFromConfig = getConfigBoundary(new Date(ClockUtils.epochMillis()));
        long lastTimestampFromMarker = nodeEventCursorService.oldestTimestampForEventLogOperator();
        long firstEventErrorTimestamp = eventErrorsService.getFirstEventErrorTimestampWithEventId();
        long minTimestamp = min(lastTimestampFromMarker, firstEventErrorTimestamp);
        long eventsMaxBoundary = max(maxBoundaryFromConfig, minTimestamp);
        return MaxBoundary.builder()
                .timestamp(eventsMaxBoundary)
                .deleteErrors(maxBoundaryFromConfig == eventsMaxBoundary)
                .build();
    }


    @VisibleForTesting
    long getConfigBoundary(Date time) {
        return new DateTime(time.getTime())
                .minusDays(ConstantValues.ageOfEventsLogEntriesToDiscardDays.getInt()).getMillis();
    }

    private void sleep() {
        try {
            Thread.sleep(ConstantValues.eventsCleanupIterationSleepTimeMillis.getLong());
        } catch (InterruptedException e) {
            Thread.currentThread().interrupt();
            throw new EventsCleanupException("Error while cleaning up events log", e);
        }
    }

    @JobCommand(jobName = JobName.EVENT_LOG_CLEANUP, singleton = true,
            schedulerUser = TaskUser.SYSTEM,
            manualUser = TaskUser.SYSTEM,
            description = "Cleanup old events from events log")
    public static class EventsLogCleanUpTask extends QuartzCommand {

        @Override
        protected void onExecute(JobExecutionContext callbackContext) throws JobExecutionException {
            ContextHelper.get().beanForType(EventsLogCleanUpService.class).cleanup();
        }
    }

    @VisibleForTesting
    @Data
    @Builder
    static class MaxBoundary {
        private long timestamp;
        private boolean deleteErrors;
    }
}
