#!/usr/bin/env bash

set -e

export PROJECT_ROOT_DIR="$(cd "$(dirname "${BASH_SOURCE[0]}")" && cd ../../.. && pwd)"
source "${PROJECT_ROOT_DIR}/build/ci/scripts/lib/commonUtils.sh"
source "${PROJECT_ROOT_DIR}/build/ci/scripts/lib/mavenUtils.sh"
source "${PROJECT_ROOT_DIR}/build/ci/scripts/lib/dockerUtils.sh"
source "${PROJECT_ROOT_DIR}/build/ci/scripts/lib/versionUtils.sh"
source "${PROJECT_ROOT_DIR}/build/ci/scripts/lib/pipelineUtils.sh"
source "${PROJECT_ROOT_DIR}/build/ci/scripts/lib/buildInfoUtils.sh"

function on_success() {
  write_BuildInfo
  export JFROG_CLI_BUILD_NUMBER="${BUILD_NUMBER}"
  export JFROG_CLI_BUILD_NAME="${ART_OSS_BUILD_NAME}"

  buildInfo_modifyBuildName "${JFROG_CLI_BUILD_NAME}" "${JFROG_CLI_BUILD_NUMBER}"
  buildInfo_publish "${JFROG_CLI_BUILD_NAME}"

  # Push and publish a build with the docker images
  pushAndUpdateResource "jfrog/artifactory-oss" "artifactory_oss_art_jfrog_io_docker_image" "$DOCKER_IMAGE_VERSION" "$DOCKER_PUSH_REPO" "$DOCKER_PUSH_REGISTRY" "$ART_OSS_DOCKER_BUILD_NAME"
  buildInfo_publish "${ART_OSS_DOCKER_BUILD_NAME}"
  return 0
}

function on_complete() {
  docker_cleanupAllUnusedContainersAndImages || echo "Docker images cleanup finished with an error"
  return 0
}

function trigger_installer_package() {
  if ! common_retry "trigger installers" 5 "__call_installer_package_hook $PRODUCT_DEPLOY_REPO"; then
    echo "[ERROR] Failed to trigger installer package pipeline." >&2
  fi
}

function __call_installer_package_hook(){
  curl -X POST "${int_installer_bundler_pipeline_hook_url}" -d \
    '{
        "PRODUCT_RESOLVE_REPO":"'${PRODUCT_DEPLOY_REPO}'",
        "PRODUCT_DEPLOY_REPO":"'${PRODUCT_DEPLOY_REPO}'",
        "PRODUCT_VERSION":"'${PRODUCT_VERSION}'",
        "PRODUCT_NAME":"artifactory_oss",
        "BUILD_NUMBER":"'${JFROG_CLI_BUILD_NUMBER}'",
        "DOCKER_PUSH_REPO": "'${DOCKER_PUSH_REPO}'",
        "INSTALLER_COMMON_REPO":"'${PRODUCT_DEPLOY_REPO}'",
        "INSTALLER_COMMON_VERSION":"'${INSTALLER_COMMON_VERSION}'",
        "DOCKER_PUSH_REGISTRY": "'${DOCKER_PUSH_REGISTRY}'",
        "ART_OSS_DOCKER_BUILD_NAME": "'${ART_OSS_DOCKER_BUILD_NAME}'",
        "DOCKER_IMAGE_VERSION" : "'${DOCKER_IMAGE_VERSION}'",
        "ART_OSS_BUILD_NAME":"'${ART_OSS_INSTALLERS_BUILD_NAME}'",
        "INSTALLER_TYPE":"'${INSTALLER_PACKAGE_TYPES}'"
    }' \
    -u "${int_installer_bundler_pipeline_hook_username}:${int_installer_bundler_pipeline_hook_password}" \
    -H 'Content-Type: application/json' \
    --silent --fail --show-error
}

function on_start() {
  cd "${PROJECT_ROOT_DIR}/artifactory-oss"
  # Publish the build with the Maven artifacts
  local version
  version="$(maven_getVersionFromPom "${PROJECT_ROOT_DIR}/artifactory-oss/pom.xml")"
  buildInfo_setBuildNameAndNumber

  local branch
  local normalizedBranch
  local repo
  echo "Calculating TARGET_VERSION and DOCKER_COMPOSE_REGISTRY for the docker compose env"
  echo "Version: ${version}"
  branch="$(pipeline_gitBranchName)"
  echo "Branch:  ${branch}"
  normalizedBranch=$(common_normalizeBranchName "${branch}")

  local targetVersion="${version}-${normalizedBranch}-${run_number}"
  local dockerComposeRegistry="docker.jfrog.io"
  if version_isRelease "${version}" || version_isMilestone "${version}"; then
    targetVersion="${version}"
    if version_isRelease "${version}"; then
      dockerComposeRegistry="docker.bintray.io"
    fi
  fi

  export TARGET_VERSION="${targetVersion}" DOCKER_COMPOSE_REGISTRY="$dockerComposeRegistry"

  # Push and publish a build with the docker images
  local repository="art-docker-dev-local"
  local imageName="jfrog/artifactory-oss"
  local dockerVirtualRegistry="docker.jfrog.io"

  PRODUCT_DEPLOY_REPO="$(maven_devArtifactRepoName "$branch")"

  PRODUCT_VERSION="$version"
  DOCKER_IMAGE_VERSION="$targetVersion"
  DOCKER_PUSH_REPO="${repository}"
  DOCKER_PUSH_REGISTRY="art.jfrog.io"

  INSTALLER_COMMON_VERSION=$(maven_getPropertyValueFromPom "jfrog.installer.commons.version" "${PROJECT_ROOT_DIR}/artifactory-oss/pom.xml")

  setCustomBuildTypes
  ART_OSS_INSTALLERS_BUILD_NAME="${JFROG_CLI_BUILD_NAME}"
  ART_OSS_DOCKER_BUILD_NAME="$(docker_generateBuildName)"
  INSTALLER_COMMON_REPO="$(maven_devArtifactRepoName "$branch")"

  export BUILD_PARAMS_JSON="${res_artifactory_oss_bitbucket_resourcePath}/installers/src/products/build-params.json"

  echo '{
        "PRODUCT_RESOLVE_REPO":"'${PRODUCT_DEPLOY_REPO}'",
        "PRODUCT_DEPLOY_REPO":"'${PRODUCT_DEPLOY_REPO}'",
        "PRODUCT_VERSION":"'${PRODUCT_VERSION}'",
        "PRODUCT_NAME":"artifactory_oss",
        "BUILD_NUMBER":"'${JFROG_CLI_BUILD_NUMBER}'",
        "DOCKER_PUSH_REPO": "'${DOCKER_PUSH_REPO}'",
        "INSTALLER_COMMON_REPO":"'${INSTALLER_COMMON_REPO}'",
        "INSTALLER_COMMON_VERSION":"'${INSTALLER_COMMON_VERSION}'",
        "DOCKER_PUSH_REGISTRY": "'${DOCKER_PUSH_REGISTRY}'",
        "ART_OSS_DOCKER_BUILD_NAME": "'${ART_OSS_DOCKER_BUILD_NAME}'",
        "DOCKER_IMAGE_VERSION" : "'${DOCKER_IMAGE_VERSION}'",
        "ART_OSS_BUILD_NAME":"'${ART_OSS_INSTALLERS_BUILD_NAME}'",
        "INSTALLER_TYPE":"'${INSTALLER_PACKAGE_TYPES}'"
    }' | jq '.' > "${BUILD_PARAMS_JSON}"

  #download installer-commons from branch snapshot
  ${res_artifactory_oss_bitbucket_resourcePath}/installers/src/products/download.sh --user ${int_entplus_deployer_user} --pass ${int_entplus_deployer_apikey}
  source ${res_artifactory_oss_bitbucket_resourcePath}/installers/src/products/build-src/configuration-manager/buildUtils/commonUtils.sh
  exportJsonParams "${BUILD_PARAMS_JSON}"

  cd ${res_artifactory_oss_bitbucket_resourcePath}/installers/src/products/build-src/configuration-manager/products/artifactory/artifactory-oss/distribution
  source ${res_artifactory_oss_bitbucket_resourcePath}/installers/src/products/build-src/configuration-manager/products/artifactory/artifactory-oss/package.sh
  source ${res_artifactory_oss_bitbucket_resourcePath}/installers/src/products/build-src/configuration-manager/products/pack_env.sh
  exportEnvAsRunVariables "${RUN_VARIABLE_KEYLIST}"

  prepare_maven_config
}

function write_BuildInfo () {
    write_output artifactory_oss_package_installers_build_info buildName="${ART_OSS_INSTALLERS_BUILD_NAME}" buildNumber="${JFROG_CLI_BUILD_NUMBER}" version="${targetVersion}"
    write_output "artifactory_oss_art_jfrog_io_docker_image" sourceRepository="${repository}" imageName="${dockerVirtualRegistry}/${imageName}" imageTag="${targetVersion}"
    write_output artifactory_oss_docker_build_info buildName="${ART_OSS_DOCKER_BUILD_NAME}" buildNumber="${JFROG_CLI_BUILD_NUMBER}"
}
function setInstallerRepo(){
    if version_isRelease "${INSTALLER_COMMON_VERSION}" || version_isMilestone "${INSTALLER_COMMON_VERSION}"; then
      INSTALLER_COMMON_REPO="libs-staging"
    else
      INSTALLER_COMMON_REPO="dev-master-maven-local"
    fi
}

# exports an environment variable with desired build types based on the last git commit body or env variable
function setCustomBuildTypes(){
  local branch
  branch="$(pipeline_gitBranchName)"
  if [[ $(common_gitLastCommitBody) == *"[package-all-types]"* ]] || [[ "${PACKAGE_ALL_TYPES}" == true ]] || common_isReleaseOrMilestoneOrMasterBranch "$branch"; then
    export INSTALLER_PACKAGE_TYPES="docker,rpm,deb"
  else
    export INSTALLER_PACKAGE_TYPES="docker"
  fi
}