package org.artifactory.storage.db;

import com.google.common.collect.Lists;
import org.apache.commons.lang3.RandomUtils;
import org.artifactory.api.context.ContextHelper;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.converter.ArtifactoryConverterAdapter;
import org.artifactory.converter.ConverterPreconditionException;
import org.artifactory.storage.db.util.JdbcHelper;
import org.artifactory.storage.db.version.ArtifactoryDBVersion;
import org.artifactory.storage.db.version.converter.AsyncDBSqlConditionalConverter;
import org.artifactory.storage.db.version.converter.DBConverter;
import org.artifactory.storage.jobs.async.database.DatabaseConverterRunnable;
import org.artifactory.util.AolUtils;
import org.artifactory.util.CollectionUtils;
import org.artifactory.version.CompoundVersionDetails;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.List;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;

/**
 * Run any necessary conversions to bring the system to the target version.
 * This converter is running as async after all the other converters were finished.
 * Need to be used only when its an optional conversion that can take time (for example adding index's).
 *
 * @author Alexei Vainshtein
 */
public class AsyncDatabaseConverterAdapter implements ArtifactoryConverterAdapter {
    private static final Logger log = LoggerFactory.getLogger(AsyncDatabaseConverterAdapter.class);
    private final ArtifactoryHome artifactoryHome;

    public AsyncDatabaseConverterAdapter(ArtifactoryHome artifactoryHome) {
        this.artifactoryHome = artifactoryHome;
    }

    @Override
    public void convert(CompoundVersionDetails source,
            CompoundVersionDetails target) {
        log.debug("Starting collection of async converters.");
        // All converters of type AsyncConverters
        List<DBConverter> converters = Lists.newArrayList();
        for (ArtifactoryDBVersion dbVersion : ArtifactoryDBVersion.values()) {
            for (DBConverter dbConverter : dbVersion.getConverters()) {
                if (dbConverter instanceof AsyncDBSqlConditionalConverter) {
                    converters.add(dbConverter);
                }
            }
        }

        // Create and run executor for this.
        if (CollectionUtils.notNullOrEmpty(converters)) {
            ScheduledExecutorService executorService = getScheduledExecutorService();
            try {
                executorService
                        .schedule(new DatabaseConverterRunnable(artifactoryHome, converters,
                                        ContextHelper.get().beanForType(JdbcHelper.class)), getMinutesDelay(),
                                TimeUnit.MINUTES);
                executorService.shutdown();
            } catch (Exception e) {
                log.error("An error occurred: {}", e.getCause(), e);
            }
        }
    }

    int getMinutesDelay() {
        int minForDelay = 0;
        if (AolUtils.isAol()) {
            minForDelay = RandomUtils.nextInt(1, 100);
            log.info("Async converter will run in {} minutes", minForDelay);
        }
        return minForDelay;
    }

    @NotNull
    protected ScheduledExecutorService getScheduledExecutorService() {
        return Executors.newScheduledThreadPool(1);
    }

    @Override
    public boolean isInterested(@Nullable CompoundVersionDetails source,
            CompoundVersionDetails target) {
        return true;
    }

    @Override
    public void revert() {

    }

    @Override
    public void backup() {

    }

    @Override
    public void clean() {

    }

    @Override
    public void assertConversionPrecondition(ArtifactoryHome home,
            CompoundVersionDetails fromVersion, CompoundVersionDetails toVersion)
            throws ConverterPreconditionException {
        if (artifactoryHome == null) {
            throw new ConverterPreconditionException("Artifactory home is missing");
        }

        if (artifactoryHome.getDBProperties() == null) {
            throw new ConverterPreconditionException("DbProperties is missing");
        }
    }
}
