package org.artifactory.rest.resource.xray;

import org.apache.commons.lang.StringUtils;
import org.apache.http.HttpEntity;
import org.apache.http.HttpStatus;
import org.apache.http.ProtocolVersion;
import org.apache.http.StatusLine;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.entity.InputStreamEntity;
import org.apache.http.message.BasicStatusLine;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.build.BuildAddon;
import org.artifactory.addon.xray.XrayAddon;
import org.artifactory.addon.xray.XrayScanBuild;
import org.artifactory.api.build.BuildService;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.projects.ProjectsService;
import org.artifactory.rest.common.model.xray.XrayScanBuildModel;
import org.artifactory.test.ArtifactoryHomeBoundTest;
import org.artifactory.util.StringInputStream;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.MockitoSession;
import org.mockito.quality.Strictness;
import org.testng.Assert;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import javax.ws.rs.core.Response;
import java.io.IOException;

import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.when;

/**
 * @author Mor Merhav
 */
public class XrayResourceTest extends ArtifactoryHomeBoundTest {

    @Mock
    private AuthorizationService authorizationServiceMock;
    @Mock
    private AddonsManager addonsManagerMock;
    @Mock
    private XrayAddon xrayAddon;
    @Mock
    private CloseableHttpResponse responseMock;
    @Mock
    private BuildService buildServiceMock;

    @InjectMocks
    private XrayResource xrayResource;
    private MockitoSession mockitoSession;

    @BeforeMethod
    public void setUp() {
        mockitoSession = Mockito.mockitoSession()
                .strictness(Strictness.STRICT_STUBS)
                .initMocks(this)
                .startMocking();
        when(addonsManagerMock.addonByType(XrayAddon.class)).thenReturn(xrayAddon);
    }

    @AfterMethod
    public void tearDown() {
        mockitoSession.finishMocking();
    }

    @Test
    public void testScanBuildDefaultBuildRepoScanBuildSuccess() throws IOException {
        String buildRepoName = BuildAddon.BUILD_INFO_REPO_NAME;
        when(buildServiceMock.getPreferredBuildRepo(buildRepoName, null))
                .thenReturn(buildRepoName);
        when(authorizationServiceMock.canReadBuild( null, null, buildRepoName))
                .thenReturn(true);
        XrayScanBuildModel xrayScanBuildModel = new XrayScanBuildModel();
        xrayScanBuildModel.setBuildRepo(BuildAddon.BUILD_INFO_REPO_NAME);
        StatusLine statusLine = new BasicStatusLine(new ProtocolVersion("proto", 1, 0), HttpStatus.SC_OK, "OK");
        when(responseMock.getStatusLine()).thenReturn(statusLine);
        HttpEntity httpEntity = new InputStreamEntity(new StringInputStream(""));
        when(responseMock.getEntity()).thenReturn(httpEntity);
        when(xrayAddon.scanBuild(XrayResourceHelper.toModel(xrayScanBuildModel))).thenReturn(responseMock);

        Response response = xrayResource.scanBuild(xrayScanBuildModel);

        Assert.assertEquals(response.getStatus(), HttpStatus.SC_OK);
    }

    @Test
    public void testScanBuildDefaultProjectRepoScanBuildSuccess() throws IOException {
        String defaultProject = ProjectsService.DEFAULT_PROJECT;
        String buildRepoName = BuildAddon.BUILD_INFO_REPO_NAME;
        when(buildServiceMock.getPreferredBuildRepo(null, defaultProject))
                .thenReturn(buildRepoName);
        when(authorizationServiceMock.canReadBuild( null, null, buildRepoName))
                .thenReturn(true);
        XrayScanBuildModel xrayScanBuildModel = new XrayScanBuildModel();
        xrayScanBuildModel.setProject(defaultProject);
        StatusLine statusLine = new BasicStatusLine(new ProtocolVersion("proto", 1, 0), HttpStatus.SC_OK, "OK");
        when(responseMock.getStatusLine()).thenReturn(statusLine);
        HttpEntity httpEntity = new InputStreamEntity(new StringInputStream(""));
        when(responseMock.getEntity()).thenReturn(httpEntity);
        XrayScanBuild xrayScanBuild = XrayResourceHelper.toModel(xrayScanBuildModel);
        xrayScanBuild.setBuildRepo(buildRepoName);
        when(xrayAddon.scanBuild(xrayScanBuild)).thenReturn(responseMock);

        Response response = xrayResource.scanBuild(xrayScanBuildModel);

        Assert.assertEquals(response.getStatus(), HttpStatus.SC_OK);
    }

    @Test
    public void testScanBuildCustomBuildRepoNoXrayProjectsSupportScanBuildFailure() {
        String buildRepoName = "custom" + BuildAddon.BUILD_INFO_REPO_EXTENSION;
        when(buildServiceMock.getPreferredBuildRepo(buildRepoName, null))
                .thenReturn(buildRepoName);
        when(authorizationServiceMock.canReadBuild( null, null, buildRepoName))
                .thenReturn(true);
        XrayScanBuildModel xrayScanBuildModel = new XrayScanBuildModel();
        xrayScanBuildModel.setBuildRepo(buildRepoName);
        when(xrayAddon.isXrayVersionSupportProjects()).thenReturn(false);

        Response response = xrayResource.scanBuild(xrayScanBuildModel);

        Assert.assertEquals(response.getStatus(), HttpStatus.SC_BAD_REQUEST);
        Assert.assertEquals(response.getEntity(),
                "Scanning non-default build repositories is unsupported by Xray version lower than: " +
                        XrayAddon.XRAY_MIN_VER_WITH_PROJECTS_SUPPORT);
    }

    @Test
    public void testScanBuildCustomProjectNoXrayProjectsSupportScanBuildFailure() {
        String projectKey = "customProject";
        String buildRepoName = projectKey + BuildAddon.BUILD_INFO_REPO_EXTENSION;
        when(buildServiceMock.getPreferredBuildRepo(null, projectKey))
                .thenReturn(buildRepoName);
        when(authorizationServiceMock.canReadBuild( null, null, buildRepoName))
                .thenReturn(true);
        XrayScanBuildModel xrayScanBuildModel = new XrayScanBuildModel();
        xrayScanBuildModel.setProject(projectKey);
        when(xrayAddon.isXrayVersionSupportProjects()).thenReturn(false);

        Response response = xrayResource.scanBuild(xrayScanBuildModel);

        Assert.assertEquals(response.getStatus(), HttpStatus.SC_BAD_REQUEST);
        Assert.assertEquals(response.getEntity(),
                "Scanning non-default build repositories is unsupported by Xray version lower than: " +
                        XrayAddon.XRAY_MIN_VER_WITH_PROJECTS_SUPPORT);
    }

    @Test
    public void testScanBuildCustomBuildRepoXrayProjectsSupportScanBuildSuccess() throws IOException {
        String buildRepoName = "custom" + BuildAddon.BUILD_INFO_REPO_EXTENSION;
        when(buildServiceMock.getPreferredBuildRepo(buildRepoName, null))
                .thenReturn(buildRepoName);
        when(authorizationServiceMock.canReadBuild( null, null, buildRepoName))
                .thenReturn(true);
        XrayScanBuildModel xrayScanBuildModel = new XrayScanBuildModel();
        xrayScanBuildModel.setBuildRepo(buildRepoName);

        StatusLine statusLine = new BasicStatusLine(new ProtocolVersion("proto", 1, 0), HttpStatus.SC_OK, "OK");
        when(responseMock.getStatusLine()).thenReturn(statusLine);
        HttpEntity httpEntity = new InputStreamEntity(new StringInputStream(""));
        when(responseMock.getEntity()).thenReturn(httpEntity);
        when(xrayAddon.isXrayVersionSupportProjects()).thenReturn(true);
        XrayScanBuild xrayScanBuild = XrayResourceHelper.toModel(xrayScanBuildModel);
        when(xrayAddon.scanBuild(xrayScanBuild)).thenReturn(responseMock);

        Response response = xrayResource.scanBuild(xrayScanBuildModel);

        Assert.assertEquals(response.getStatus(), HttpStatus.SC_OK);
    }

    @Test
    public void testScanBuildCustomProjectXrayProjectsSupportScanBuildSuccess() throws IOException {
        String projectKey = "customProject";
        String buildRepoName = projectKey + BuildAddon.BUILD_INFO_REPO_EXTENSION;
        when(buildServiceMock.getPreferredBuildRepo(null, projectKey))
                .thenReturn(buildRepoName);
        when(authorizationServiceMock.canReadBuild( null, null, buildRepoName))
                .thenReturn(true);
        XrayScanBuildModel xrayScanBuildModel = new XrayScanBuildModel();
        xrayScanBuildModel.setProject(projectKey);

        StatusLine statusLine = new BasicStatusLine(new ProtocolVersion("proto", 1, 0), HttpStatus.SC_OK, "OK");
        when(responseMock.getStatusLine()).thenReturn(statusLine);
        HttpEntity httpEntity = new InputStreamEntity(new StringInputStream(""));
        when(responseMock.getEntity()).thenReturn(httpEntity);
        when(xrayAddon.isXrayVersionSupportProjects()).thenReturn(true);
        XrayScanBuild xrayScanBuild = XrayResourceHelper.toModel(xrayScanBuildModel);
        xrayScanBuild.setBuildRepo(buildRepoName);
        when(xrayAddon.scanBuild(xrayScanBuild)).thenReturn(responseMock);

        Response response = xrayResource.scanBuild(xrayScanBuildModel);

        Assert.assertEquals(response.getStatus(), HttpStatus.SC_OK);
    }

    @Test
    public void testScanBuildBlankProjectBlankBuildRepoDefaultBuildRepoSuccess() throws IOException {
        String buildRepoName = BuildAddon.BUILD_INFO_REPO_NAME;
        when(authorizationServiceMock.canReadBuild( null, null, buildRepoName))
                .thenReturn(true);
        when(buildServiceMock.getPreferredBuildRepo(any(), any())).thenReturn(buildRepoName);

        XrayScanBuildModel xrayScanBuildModel = new XrayScanBuildModel();
        xrayScanBuildModel.setProject(null);
        xrayScanBuildModel.setBuildRepo(buildRepoName);

        StatusLine statusLine = new BasicStatusLine(new ProtocolVersion("proto", 1, 0), HttpStatus.SC_OK, "OK");
        when(responseMock.getStatusLine()).thenReturn(statusLine);
        HttpEntity httpEntity = new InputStreamEntity(new StringInputStream(""));
        when(responseMock.getEntity()).thenReturn(httpEntity);
        XrayScanBuild xrayScanBuild = XrayResourceHelper.toModel(xrayScanBuildModel);
        when(xrayAddon.scanBuild(xrayScanBuild)).thenReturn(responseMock);

        Response response = xrayResource.scanBuild(xrayScanBuildModel);

        Assert.assertEquals(response.getStatus(), HttpStatus.SC_OK);

        xrayScanBuildModel.setProject(StringUtils.EMPTY);
        xrayScanBuildModel.setBuildRepo(StringUtils.EMPTY);

        response = xrayResource.scanBuild(xrayScanBuildModel);

        Assert.assertEquals(response.getStatus(), HttpStatus.SC_OK);
    }
}